module Sol_Exercise_10 where
import Data.Char (isSpace, isDigit, digitToInt)
import Data.List (find, isInfixOf, isSuffixOf, nub, sort, intersect, inits, sortBy, partition, genericLength)
import Data.Function (on)
import Network
import System.Environment (getArgs, withArgs)
import System.Random (randomRIO)
import System.IO
import Test.QuickCheck
import Text.Read (readMaybe)

import Util_10

{- Library - nicht veraendern -}

type Point = (Rational, Rational)
type Polygon = [Point]
data City = City String Point Integer deriving Show
data District = District Polygon [City] deriving Show

{- Ende Library -}

{- G1 -}

{- Using global generator -}
nRandomR :: (Int, Int) -> Int -> IO [Int]
nRandomR lowhigh = randHelp []
  where
    randHelp :: [Int] -> Int -> IO [Int]
    randHelp xs 0 = return xs
    randHelp xs n = do
      i <- randomRIO lowhigh
      if i `elem` xs then randHelp xs n else randHelp (i : xs) (n - 1)

{-
{- Using local generator (needs to be threaded through) -}
nRandomR' :: (Int, Int) -> Int -> IO [Int]
nRandomR' lowhigh n = do
  gen <- newStdGen
  return (randHelp gen [] n)
  where
    randHelp :: StdGen -> [Int] -> Int -> [Int]
    randHelp gen xs 0 = xs
    randHelp gen xs n =
      if i `elem` xs then randHelp gen' xs n else randHelp gen' (i : xs) (n - 1)
      where (i, gen') = randomR (0,9) gen
-}


{- G2 -}

getLineInt :: IO Int
getLineInt = do
    line <- getLine
    case readMaybe line :: Maybe Int of
        Nothing -> do
            putStrLn "Not a number"
            getLineInt
        Just n -> return n

guessNum :: IO Int
guessNum = do
    rnd <- randomRIO (0,100)
    putStrLn "Guess a number between 0 and 100"
    doGuessNum rnd 1
  where
    doGuessNum rnd cnt = do
      num <- getLineInt
      if num < rnd then do
        putStrLn "The number you are looking for is greater"
        doGuessNum rnd (cnt+1)
      else if num > rnd then do
        putStrLn "The number you are looking for is smaller"
        doGuessNum rnd (cnt+1)
      else do
        putStrLn "You found it!"
        return cnt

{- H1 -}

countNumbers :: IO Int
countNumbers = count 0
  where
    count n = do
        l <- getLine
        if null l
        then return n
        else count (next n l)
    next n l = case (readMaybe l :: Maybe Int) of
        Just _ -> 1
        Nothing -> 0

{- H2 -}

joinFiles :: FilePath -> FilePath -> IO [String]
joinFiles namePath exPath = do
    names <- readFile namePath
    ex <- readFile exPath
    return $ doJoin (lines names) (lines ex)

doJoin :: [String] -> [String] -> [String]
doJoin (x:xs) (y:ys)
    | ix == iy = (ix ++ rx ++ ry) : doJoin xs ys
    | ix < iy = doJoin xs (y:ys)
    | ix > iy = doJoin (x:xs) ys
  where
    (ix,rx) = break (==',') x
    (iy,ry) = break (==',') y
doJoin _ _ = []

main :: IO ()
main = do
    [namePath, exPath] <- getArgs
    join <- joinFiles namePath exPath
    putStrLn $ unlines join

{- H3 -}

client :: IO ()
client = withSocketsDo $ do
  h <- connectTo "fp.in.tum.de" $ PortNumber 8080
  _ <- hGetLine h
  hPutStrLn h "lars@example.org 42"
  values <- hGetLine h
  case words values of
    [a, b] | Just a' <- readMaybe a, Just b' <- readMaybe b -> do
      hPutStrLn h $ "I computed " ++ show (42 ^ a' - b') ++ ", is that right?"
      reply <- hGetLine h
      putStrLn reply
    _ ->
      putStrLn "OH NOES!!!"
  hClose h

{- H4 -}

naiveSplit :: District -> Integer -> [Polygon]
naiveSplit (District border _) = partition border
  where partition ps 1 = [ps]
        partition (p1 : p2 : p3 : ps) n = [p1, p2, p3] : partition (p1 : p3 : ps) (n - 1)
        partition _ _ = error ""

cityInPolygon :: Polygon -> City -> Bool
cityInPolygon p (City _ pos _) = inPolygon pos p

allSplits :: Polygon -> (Integer, Integer) -> [(Polygon, Polygon)]
allSplits (p1 : ps) (min1, min2) = all [] p1 ps
  where all init start (skip : rest) =
          [ (left, right) | ps <- tail $ inits rest
                          , let rem = length rest - length ps
                          , length init + rem > 0
                          , let left = [start, skip] ++ ps
                          , let right = [start] ++ drop (length ps - 1) rest ++ init
                          , genericLength left >= min1
                          , genericLength right >= min2
                          ] ++
          all (init ++ [start]) skip rest
        all _ _ [] = []
allSplits [] _ = error "allSplits: not a valid polygon"

split :: District -> Integer -> [Polygon]
split (District border cities) regions = splitInner regions cities border
  where splitInner 1 _ poly = [poly]
        splitInner n cities poly | n > 1 = splitInner n1 cities1 poly1 ++ splitInner n2 cities2 poly2
          where n1 = n `div` 2
                n2 = n - n1
                inhabitants (City _ _ i) = i
                totalInhabitants = sum $ map inhabitants cities
                target = (totalInhabitants * n1) `div` n

                mkSplit (poly1, poly2) = (diff, (poly1, cities1, poly2, cities2))
                  where (cities1, cities2) = partition (cityInPolygon poly1) cities
                        diff = abs $ target - sum (map inhabitants cities1)

                (poly1, cities1, poly2, cities2) = snd $ head $ sortBy (compare `on` fst) $ map mkSplit $ allSplits poly (n1 + 2, n2 + 2)

        splitInner _ _ _ = error "split: split number too small"
